/*
 *  CPE 471 lab 4 - modern graphics test bed
 *  draws a partial cube using a VBO and IBO 
 *  glut/OpenGL/GLSL application   
 *  Uses glm and local matrix stack
 *  to handle matrix transforms for a view matrix, projection matrix and
 *  model transform matrix
 *
 *  zwood 9/12 
 *  Copyright 2012 Cal Poly. All rights reserved.
 *
 *****************************************************************************/
#include "gl.h"

#include <iostream>
#include <vector>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include "glm/glm.hpp"
#include "glm/gtc/matrix_transform.hpp"
#include "glm/gtc/type_ptr.hpp"
#include "GLSL_helper.h"
#include "CameraHelper.h"
#include "ModelHelper.h"
#include "MStackHelp.h"
#include "GeometryCreator.h"
#include "PhysicsEngine.h"
#include "types.h"

#define FORWARD_SPEED 0.05
#define BACKWARD_SPEED 0.025
#define STRAFE_SPEED 0.05
#define PITCH_SPEED 0.01
#define YAW_SPEED 0.01
#define FLY_SPEED 0.01

#define LIGHT_RADIUS 4.0
#define STEP_SIZE 10
#define SUN_SPEED .01f

#define CAMERA_HEIGHT 0.2

#define ROOM_SIZE 5

int ShadeProg;

//Handles to the shader data
struct glHandles h;

/* globals to control positioning and window size */
float g_width, g_height;

bool g_verbose = false;
bool g_night = false;

bool g_mouseClick = false;
int g_mouseClickX = 0;
int g_mouseClickY = 0;

CameraHelper Camera;
std::vector<ModelHelper> Models;
RenderingHelper ModelTrans;
ModelHelper LightHelper;
PhysicsEngine* Engine;

void setRoomProperties(ModelHelper & model) {
    model.specularCoefficient = 1.0;
    model.diffuseCoefficient = 1.0;
    model.fixed = true;
    model.sphere = false;
}

void InitGeom() {
    vec3 roomColor = vec3(.1, .5, .1);

    Mesh *top = GeometryCreator::CreateCube(vec3(ROOM_SIZE, 0, ROOM_SIZE));
    Mesh *side = GeometryCreator::CreateCube(vec3(ROOM_SIZE, ROOM_SIZE, 0));

    ModelHelper ground = ModelHelper(top, roomColor);
    setRoomProperties(ground);
    ground.n = vec3(0,1,0);

    ModelHelper left = ModelHelper(side, roomColor);
    setRoomProperties(left);
    left.addTranslate(-ROOM_SIZE / 2.0, ROOM_SIZE / 2.0, 0);
    left.addRotation(vec3(0, 1, 0), 90);
    left.n = vec3(1,0,0);

    ModelHelper right = ModelHelper(side, roomColor);
    setRoomProperties(right);
    right.addTranslate(ROOM_SIZE / 2.0, ROOM_SIZE / 2.0, 0);
    right.addRotation(vec3(0, 1, 0), 90);
    right.n = vec3(-1,0,0);
    
    ModelHelper forward = ModelHelper(side, roomColor);
    setRoomProperties(forward);
    forward.addTranslate(0, ROOM_SIZE / 2.0, ROOM_SIZE / 2.0);
    forward.n = vec3(0,0,-1);

    ModelHelper back = ModelHelper(side, roomColor);
    setRoomProperties(back);
    back.addTranslate(0, ROOM_SIZE / 2.0, -ROOM_SIZE / 2.0);
    back.n = vec3(0,0,1);

    ModelHelper ceiling = ModelHelper(top, roomColor);
    setRoomProperties(ceiling);
    ceiling.addTranslate(0, ROOM_SIZE, 0);
    ceiling.n = vec3(0,-1,0);

    Models.push_back(ground);
    Models.push_back(left);
    Models.push_back(right);
    Models.push_back(forward);
    Models.push_back(back);
    Models.push_back(ceiling);
}

void SetProjectionMatrix() {
    glm::mat4 Projection = glm::perspective(80.0f, (float) g_width / g_height, 0.1f, 100.f);
    safe_glUniformMatrix4fv(h.uProjMatrix, glm::value_ptr(Projection));
}

void SetView() {
    vec3 eye = Camera.getEye();
    vec3 center = Camera.getCenter();
    glm::mat4 View = glm::lookAt(eye, center, vec3(0.0, 1.0, 0.0));
    safe_glUniformMatrix4fv(h.uViewMatrix, glm::value_ptr(View));
}

void SetLight() {
    // rotate about x but stay 100 away
    vec3 offset = LightHelper.offset;
    mat4 rot = LightHelper.rotMatrix;
    ModelTrans.useModelViewMatrix();
    ModelTrans.loadIdentity();
    ModelTrans.pushMatrix();
        ModelTrans.multMatrix(rot);
        ModelTrans.translate(offset);
        vec4 pos = ModelTrans.modelViewMatrix * vec4(0,0,0,1.0);
    ModelTrans.popMatrix();
    g_night = pos.y < 0.0;
    safe_glUniform3f(h.uLight, pos.x, pos.y, pos.z);
    
    float skyval = std::min(1.0f, std::max((pos.y * 0.01f) + 1, 0.5f));
    glClearColor(0.6f * skyval, 0.6f * skyval, 1.0f * skyval, 1.0f);
}

void SetCamera() {
    vec3 eye = Camera.getEye();
    safe_glUniform3f(h.uCamera, eye.x, eye.y, eye.z);
}

void InitShaderHandles(int ShadeProg) {
    h.aPosition = safe_glGetAttribLocation(ShadeProg, "aPosition");
    h.aNormal = safe_glGetAttribLocation(ShadeProg, "aNormal");

    h.uColor = safe_glGetUniformLocation(ShadeProg, "uColor");
    h.uDiffuse = safe_glGetUniformLocation(ShadeProg, "uDiffuseCoefficient");
    h.uSpecular = safe_glGetUniformLocation(ShadeProg, "uSpecularCoefficient");
    h.uSpecularHardness = safe_glGetUniformLocation(ShadeProg, "uSpecularHardness");
    h.uModelMatrix = safe_glGetUniformLocation(ShadeProg, "uModelMatrix");
    h.uNormalMatrix = safe_glGetUniformLocation(ShadeProg, "uNormalMatrix");

    h.uLight = safe_glGetUniformLocation(ShadeProg, "uLight");
    h.uCamera = safe_glGetUniformLocation(ShadeProg, "uCamera");
    h.uProjMatrix = safe_glGetUniformLocation(ShadeProg, "uProjMatrix");
    h.uViewMatrix = safe_glGetUniformLocation(ShadeProg, "uViewMatrix");
}

/*function to help load the shaders (both vertex and fragment */
int InstallShader(const GLchar *vShaderName, const GLchar *fShaderName) {
    GLuint VS; //handles to shader object
    GLuint FS; //handles to frag shader object
    GLint vCompiled, fCompiled, linked; //status of shader

    VS = glCreateShader(GL_VERTEX_SHADER);
    FS = glCreateShader(GL_FRAGMENT_SHADER);

    //load the source
    glShaderSource(VS, 1, &vShaderName, NULL);
    glShaderSource(FS, 1, &fShaderName, NULL);

    //compile shader and print log
    glCompileShader(VS);
    /* check shader status requires helper functions */printOpenGLError();
    glGetShaderiv(VS, GL_COMPILE_STATUS, &vCompiled);
    printShaderInfoLog(VS);

    //compile shader and print log
    glCompileShader(FS);
    /* check shader status requires helper functions */printOpenGLError();
    glGetShaderiv(FS, GL_COMPILE_STATUS, &fCompiled);
    printShaderInfoLog(FS);

    if (!vCompiled || !fCompiled) {
        printf("Error compiling either shader %s or %s", vShaderName, fShaderName);
        return 0;
    }

    //create a program object and attach the compiled shader
    ShadeProg = glCreateProgram();
    glAttachShader(ShadeProg, VS);
    glAttachShader(ShadeProg, FS);

    glLinkProgram(ShadeProg);
    /* check shader status requires helper functions */printOpenGLError();
    glGetProgramiv(ShadeProg, GL_LINK_STATUS, &linked);
    printProgramInfoLog(ShadeProg);

    glUseProgram(ShadeProg);

    InitShaderHandles(ShadeProg);
    printf("successfully installed shader %d\n", ShadeProg);
    return 1;
}

/* Some OpenGL initialization */
void Initialize() {
    glClearColor(0.8f, 0.8f, 1.0f, 1.0f);

    glClearDepth(1.0f);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);

    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    ModelTrans.useModelViewMatrix();
    ModelTrans.loadIdentity();

    Camera = CameraHelper();
    Camera.fly(CAMERA_HEIGHT);

    LightHelper = ModelHelper();
    LightHelper.offset = vec3(0,LIGHT_RADIUS,0); // noon

    Engine = new PhysicsEngine(Models, h);
}

/* Main display function */
void Draw(void) {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    //Start our shader
    glUseProgram(ShadeProg);

    /* only set the projection and view matrix once */
    SetProjectionMatrix();
    SetView();
    SetLight();
    SetCamera();

    for (std::vector<ModelHelper>::iterator it = Models.begin(); it != Models.end(); ++it) {
        ModelTrans.loadIdentity();
        it->draw(ModelTrans, h);
    }

    //disable the shader
    glUseProgram(0);
    glutSwapBuffers();
    printOpenGLError();
}

/* Reshape */
void ReshapeGL(int width, int height) {
    g_width = (float) width;
    g_height = (float) height;
    glViewport(0, 0, (GLsizei) (width), (GLsizei) (height));
}

void addSphere(vec3 color, float r, float mass, float elasticity) {
    ModelHelper sphere = ModelHelper(GeometryCreator::CreateSphere(vec3(r, r, r)),
            color);
    sphere.addTranslate(Camera.getCenter());
    sphere.r = r; // TODO: Make this dynamic
    sphere.mass = mass;
    sphere.elasticity = elasticity;
    Models.push_back(sphere);
}

//the keyboard callback to change the values to the transforms
void keyboard(unsigned char key, int x, int y) {
    switch (key) {
    /* WASD keyes effect view/camera transform */
    case 'w':
        Camera.move(FORWARD_SPEED);
        break;
    case 's':
        Camera.move(-BACKWARD_SPEED);
        break;
    case 'a':
        Camera.strafe(STRAFE_SPEED);
        break;
    case 'd':
        Camera.strafe(-STRAFE_SPEED);
        break;
    case 'z':
        Camera.fly(-FLY_SPEED);
        break;
    case 'x':
        Camera.fly(FLY_SPEED);
        break;
    case 'r':
	Engine->gravity = vec3(0, GRAVITY, 0);
        Camera.reset();
        Camera.fly(CAMERA_HEIGHT);
        Models.clear();
        InitGeom();
        break;
    case 'c':
        addSphere(vec3(0.1,0.1,0.8), 0.1, 1.0, 0.8);
        break;
    case 'v':
        addSphere(vec3(0.8,0.1,0.1), 0.15, 1.0, 0.8);
        break;
    case 'b':
        addSphere(vec3(0.1,0.4,0.8), 0.1, 2.0, 0.8);
        break;
    case 'n':
        addSphere(vec3(0.8,0.4,0.1), 0.15, 2.0, 0.8);
        break;
    case 'm':
        addSphere(vec3(0.8,0.8,0.8), 0.1, 1.0, 1.0);
        break;
    case ',':
        addSphere(vec3(0.1,0.1,0.1), 0.075, 10.0, 0.8);
	break;
    case 't':
	Engine->gravity = vec3(0, GRAVITY, 0);
	break;
    case 'y':
	Engine->gravity = vec3(0, -1.0 * GRAVITY, 0);
	break;
    case 'u':
	Engine->gravity = vec3(GRAVITY, 0, 0);
	break;
    case 'i':
	Engine->gravity = vec3(-1.0 * GRAVITY, 0, 0);
	break;
    case 'o':
	Engine->gravity = vec3(0, 0, GRAVITY);
	break;
    case 'p':
	Engine->gravity = vec3(0, 0, -1.0 * GRAVITY);
	break;
    case 'q':
    case 'Q':
        exit(EXIT_SUCCESS);
        break;
    }
}

void mouse(int button, int state, int x, int y) {
    unsigned char color[3];

    if (button == GLUT_LEFT_BUTTON) {
        if (state == GLUT_DOWN) {
            g_mouseClickX = x;
            g_mouseClickY = y;
            g_mouseClick = true;
        } else {
            g_mouseClick = false;
        }
    }
}

//the mouse move callback
void mouseMove(int x, int y) {
    if (g_mouseClick) {
        int moveX = x - g_mouseClickX; // silly coordinates
        int moveY = g_mouseClickY - y;

        Camera.addYaw(moveX * YAW_SPEED);
        Camera.addPitch(moveY * PITCH_SPEED);

        g_mouseClickX = x;
        g_mouseClickY = y;
    }
}

void Timer(int param)
{
    Engine->tick(STEP_SIZE);
    //LightHelper.addRotation(vec3(0,0,1), STEP_SIZE * SUN_SPEED * (g_night ? 2.0 : 1.0));
    glutPostRedisplay();
    glutTimerFunc(STEP_SIZE, Timer, 1);
}

int main(int argc, char *argv[]) {
    int width = 600;
    int height = 600;
    if (argc != 1 && (argc != 2 || strcmp(argv[1], "-v") != 0)) {
        std::cerr << "Usage: scene [-v]" << std::endl;
        return 1;
    }
    g_verbose = (argc == 2);

    glutInit(&argc, argv);
    glutInitWindowPosition(20, 20);
    glutInitWindowSize(width, height);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
    glutCreateWindow("World Scene");
    glutReshapeFunc(ReshapeGL);
    glutDisplayFunc(Draw);
    glutKeyboardFunc(keyboard);
    glutMouseFunc(mouse);
    glutMotionFunc(mouseMove);
    glutTimerFunc(STEP_SIZE, Timer, 1);

    // GLEW Setup (Windows only)
#ifdef _WIN32
    GLenum err = glewInit();
    if (GLEW_OK != err)
    {
        std::cerr << "Error initializing glew! " << glewGetErrorString(err) << std::endl;
        return 1;
    }
#endif

    Initialize();

    //test the openGL version
    getGLversion();
    //install the shader
    if (!InstallShader(textFileRead((char *) "phong_vert.glsl"),
            textFileRead((char *) "phong_frag.glsl"))) {
        printf("Error installing shader!\n");
        return 0;
    }

    InitGeom();
    glutMainLoop();
    return 0;
}
