/*
 * ModelHelper.cpp
 *
 *  Created on: Oct 18, 2012
 *      Author: Karl
 */

#include "ModelHelper.h"

void ModelHelper::init() {
   fixed = false;
   sphere = true;
   diffuseCoefficient = 0.5;
   specularCoefficient = 2.0;
   specularHardness = 1.0;
   elasticity = 0.8;
   mass = 1.0;
   reset();
}

void ModelHelper::reset() {
   offset = vec3(0);
   xScale = 1.0;
   yScale = 1.0;
   zScale = 1.0;
   rotMatrix = mat4(1);

   velocity = vec3(0.0);
   recalculateBB();
}

void ModelHelper::addTranslate(vec3 t) {
   addTranslate(t.x, t.y, t.z);
}

void ModelHelper::addTranslate(float x, float y, float z) {
   offset.x = std::min(MAX_TRANS, std::max(MIN_TRANS, offset.x + x));
   offset.y = std::min(MAX_TRANS, std::max(MIN_TRANS, offset.y + y));
   offset.z = std::min(MAX_TRANS, std::max(MIN_TRANS, offset.z + z));
   recalculateBB();
}

void ModelHelper::addScale(float x, float y, float z) {
   xScale = std::min(MAX_SCALE, std::max(MIN_SCALE, xScale + x));
   yScale = std::min(MAX_SCALE, std::max(MIN_SCALE, yScale + y));
   zScale = std::min(MAX_SCALE, std::max(MIN_SCALE, zScale + z));
   recalculateBB();
}

void ModelHelper::addRotation(vec3 axis, float angle) {
   if (isnan(angle) || isnan(axis.x) || isnan(axis.y) || isnan(axis.z)
         || (axis.x == 0.0 && axis.y == 0.0 && axis.z == 0.0)) {
      printf("Caught bad rotate\n");
   } else {
      rotMatrix = glm::rotate(angle, axis) * rotMatrix;
      recalculateBB();
   }
}

void ModelHelper::addVelocity(vec3 t) {
   addVelocity(t.x, t.y, t.z);
}

void ModelHelper::addVelocity(float x, float y, float z) {
   if (!fixed) {
      velocity.x = std::min(MAX_VEL, std::max(MIN_VEL, velocity.x + x));
      velocity.y = std::min(MAX_VEL, std::max(MIN_VEL, velocity.y + y));
      velocity.z = std::min(MAX_VEL, std::max(MIN_VEL, velocity.z + z));
   }
}

void ModelHelper::tick(int millis) {
   float move = (millis / 1000.0);
   addTranslate(velocity.x * move, velocity.y * move, velocity.z * move);
}

// technically, point should be a intersection manifold, but I'm not doing
// full rigid body dynamics yet
// also doesn't handle heirarchically modeled stuff
bool ModelHelper::intersect(ModelHelper & o, vec3 * myIntersect, vec3 * oIntersect, vec3 * normal) {
   if (sphere && o.sphere) {
      // check distance between centers
      glm::float_t d = glm::distance(offset, o.offset);
      if (d < (r + o.r)) {
         vec3 dirFromO = (o.offset - offset) / d;
         *myIntersect = dirFromO * (glm::float_t)r;
         *oIntersect = -dirFromO * (glm::float_t)o.r;
         *normal = dirFromO;
         return true;
      }
   } else if (sphere && !o.sphere) {
      vec3 p = o.offset - offset;
      float d = -o.n.x * p.x + -o.n.y * p.y + -o.n.z * p.z;
      if (d <= r && d >= -r) {
         *myIntersect = r * -o.n;
         *oIntersect = d * o.n;
         *normal = -o.n;
         return true;
      }
   } else if (!sphere && o.sphere) {
      vec3 p = offset - o.offset;
      float d = -n.x * p.x + -n.y * p.y + -n.z * p.z;
      if (d <= o.r && d >= -o.r) {
         *myIntersect = d * n;
         *oIntersect = o.r * -n;
         *normal = n;
         return true;
      }
   }
   return false;
}

void ModelHelper::recalculateBB() {
   // TODO
}

void ModelHelper::draw(RenderingHelper & ModelTrans, struct glHandles & h) {
   ModelTrans.pushMatrix();
   ModelTrans.translate(offset);
   ModelTrans.multMatrix(rotMatrix);
   ModelTrans.translate(-attachment);

   ModelTrans.scale(xScale, yScale, zScale); // hmm

   for (std::vector<ModelHelper*>::iterator it = children.begin(); it != children.end(); ++it) {
      (*it)->draw(ModelTrans, h);
   }

   safe_glUniformMatrix4fv(h.uModelMatrix, glm::value_ptr(ModelTrans.modelViewMatrix));
   safe_glUniformMatrix4fv(h.uNormalMatrix,
         glm::value_ptr(glm::transpose(glm::inverse(ModelTrans.modelViewMatrix))));

   safe_glEnableVertexAttribArray(h.aPosition);
   glBindBuffer(GL_ARRAY_BUFFER, mesh->PositionHandle);
   safe_glVertexAttribPointer(h.aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);

   safe_glEnableVertexAttribArray(h.aNormal);
   glBindBuffer(GL_ARRAY_BUFFER, mesh->NormalHandle);
   safe_glVertexAttribPointer(h.aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);

   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mesh->IndexHandle);

   glUniform3f(h.uColor, color.x, color.y, color.z);
   safe_glUniform1f(h.uDiffuse, diffuseCoefficient);
   safe_glUniform1f(h.uSpecular, specularCoefficient);
   safe_glUniform1f(h.uSpecularHardness, specularHardness);

   glDrawElements(GL_TRIANGLES, mesh->IndexBufferLength, GL_UNSIGNED_SHORT, 0);

   safe_glDisableVertexAttribArray(h.aPosition);
   safe_glDisableVertexAttribArray(h.aNormal);

   ModelTrans.popMatrix();
}
